use crate::Asset;
use serde::{Serialize, Serializer};

pub type AmountCents = i128;

/// Amount
///
/// The cents are stored in their lowest denomination, or their "cents".  For
/// instance a Bitcoin would be represented with a `precision` of 8, or dollars
/// in a 2 or 3.
///
/// For instance, if the dollar Asset has a `precision` of `2`, an `cents` of
/// 1255 will represent `12.55 USD`.
///
/// Amount will abstract all the math operations that can be performed with the
/// amounts, guarding that the `Asset` are the same before performing any
/// operation.
///
///
/// The `cents` and `Asset.id` must be used to store amounts in the storage
/// layer. Float or string representations should be used to display
#[derive(Copy, Clone, Debug, Eq, PartialEq)]
pub struct Amount {
    asset: Asset,
    cents: AmountCents,
}

impl Serialize for Amount {
    fn serialize<S>(&self, serializer: S) -> Result<S::Ok, S::Error>
    where
        S: Serializer,
    {
        let serialized = self.to_string();
        serializer.serialize_str(&serialized)
    }
}

impl Amount {
    pub fn new(asset: Asset, cents: AmountCents) -> Self {
        Self { asset, cents }
    }

    #[inline]
    pub fn asset(&self) -> &Asset {
        &self.asset
    }

    #[inline]
    pub fn cents(&self) -> AmountCents {
        self.cents
    }

    pub fn checked_add(&self, other: &Self) -> Option<Self> {
        if self.asset != other.asset {
            return None;
        }

        self.cents.checked_add(other.cents).map(|cents| Self {
            asset: self.asset,
            cents,
        })
    }
}

impl ToString for Amount {
    fn to_string(&self) -> String {
        let str = self.cents.to_string();
        let precision: usize = self.asset.precision.into();
        let str = if str.len() < precision + 1 {
            format!("{}{}", "0".repeat(precision - str.len() + 1), str)
        } else {
            str
        };

        let (left, right) = str.split_at(str.len() - precision);
        format!("{}.{}", left, right)
    }
}

#[cfg(test)]
mod test {
    use super::*;

    #[test]
    fn dollar() {
        let usd = Asset {
            id: 1,
            precision: 4,
        };
        let amount = usd.new_amount(1022100);
        assert_eq!(amount.to_string(), "102.2100");
    }

    #[test]
    fn bitcoin() {
        let btc = Asset {
            id: 1,
            precision: 8,
        };
        assert_eq!(btc.new_amount(1022100).to_string(), "0.01022100");
        assert_eq!(btc.new_amount(10).to_string(), "0.00000010");
        assert_eq!(btc.new_amount(10000000).to_string(), "0.10000000");
        assert_eq!(btc.new_amount(100000000).to_string(), "1.00000000");
        assert_eq!(
            btc.new_amount(100000000)
                .checked_add(&btc.new_amount(100000000))
                .unwrap()
                .to_string(),
            "2.00000000"
        );
    }
}
