use crate::{
    amount::AmountCents, asset::AssetId, transaction::from_db, AccountId, Amount, Payment,
    PaymentId, Status, Transaction, TransactionId,
};

#[derive(thiserror::Error, Debug)]
pub enum Error {
    #[error("Storage error: {0}")]
    Storage(String),

    #[error("Spend payment: {0}")]
    SpendPayment(String),

    #[error("No storage update when expecting")]
    NoUpdate,

    #[error("Record not found")]
    NotFound,

    /// TODO: Convert the AmountCents to Amount for better error reporting upstream
    #[error("Not enough unspent payments (missing {0} cents)")]
    NotEnoughUnspentPayments(AmountCents),
}

#[async_trait::async_trait]
pub trait Batch<'a> {
    async fn spend_payment(
        &mut self,
        payment_id: &PaymentId,
        status: Status,
        transaction_id: &TransactionId,
    ) -> Result<(), Error>;

    async fn rollback(self) -> Result<(), Error>;

    async fn get_payment_status(
        &mut self,
        transaction_id: &TransactionId,
    ) -> Result<Option<Status>, Error>;

    async fn commit(self) -> Result<(), Error>;

    async fn store_new_payments(&mut self, outputs: &[Payment]) -> Result<(), Error>;

    async fn store_transaction(&mut self, transaction: &Transaction) -> Result<(), Error>;
}

#[async_trait::async_trait]
pub trait Storage<'a, B>
where
    B: Batch<'a>,
{
    async fn begin(&'a self) -> Result<B, Error>;

    async fn get_payment(&self, id: PaymentId) -> Result<Payment, Error>;

    async fn get_unspent_payment(&self, id: PaymentId) -> Result<Payment, Error> {
        let payment = self.get_payment(id).await?;
        if payment.spent_by.is_some() {
            Err(Error::NotFound)
        } else {
            Ok(payment)
        }
    }

    async fn get_balance(&self, account: &AccountId) -> Result<Vec<Amount>, Error>;

    async fn get_unspent_payments(
        &self,
        account: &AccountId,
        asset: AssetId,
        target_amount: AmountCents,
    ) -> Result<Vec<Payment>, Error>;

    async fn get_transaction(
        &self,
        transaction_id: &TransactionId,
    ) -> Result<from_db::Transaction, Error>;
}
